/*
   AngelCode Tool Box Library
   Copyright (c) 2007 Andreas Jnsson
  
   This software is provided 'as-is', without any express or implied 
   warranty. In no event will the authors be held liable for any 
   damages arising from the use of this software.

   Permission is granted to anyone to use this software for any 
   purpose, including commercial applications, and to alter it and 
   redistribute it freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you 
      must not claim that you wrote the original software. If you use
      this software in a product, an acknowledgment in the product 
      documentation would be appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and 
      must not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source 
      distribution.
  
   Andreas Jnsson
   andreas@angelcode.com
*/

#include "acgfx_font.h"
#include "acgfx_dynrender.h"
#include "acutil_log.h"

using namespace std;

namespace acGraphics
{

CFont::CFont()
{
	fontHeight = 0;
	base = 0;
	scaleW = 0;
	scaleH = 0;
	scale = 1.0f;
	render = 0;

	for( int n = 0; n < 256; n++ )
		chars[n] = 0;
}

CFont::~CFont()
{
	UINT n;
	for( n = 0; n < 256; n++ )
		if( chars[n] )
			delete chars[n];

	for( n = 0; n < pages.size(); n++ )
		if( pages[n] )
			pages[n]->Release();
}

int CFont::Init(const char *fontFile, CDynRender *render)
{
	this->render = render;

	FILE *f = fopen(fontFile, "rb");

	string line;

	while( !feof(f) )
	{
		// Read until line feed (or EOF)
		line = "";
		line.reserve(256);
		while( !feof(f) )
		{
			char ch;
			if( fread(&ch, 1, 1, f) )
			{
				if( ch != '\n' ) 
					line += ch; 
				else
					break;
			}
		}

		// Skip white spaces
		int pos = SkipWhiteSpace(line, 0);

		// Read token
		int pos2 = FindEndOfToken(line, pos);
		string token = line.substr(pos, pos2-pos);

		// Interpret line
		if( token == "common" )
			InterpretCommon(line, pos2);
		else if( token == "char" )
			InterpretChar(line, pos2);
		else if( token == "kerning" )
			InterpretKerning(line, pos2);
		else if( token == "page" )
			InterpretPage(line, pos2, fontFile);
	}

	fclose(f);

	return 0;
}

int CFont::SkipWhiteSpace(string &str, int start)
{
	UINT n = start;
	while( n < str.size() )
	{
		char ch = str[n];
		if( ch != ' ' && 
			ch != '\t' && 
			ch != '\r' && 
			ch != '\n' )
			break;

		++n;
	}

	return n;
}

int CFont::FindEndOfToken(string &str, int start)
{
	UINT n = start;
	if( str[n] == '"' )
	{
		n++;
		while( n < str.size() )
		{
			char ch = str[n];
			if( ch == '"' )
			{
				// Include the last quote char in the token
				++n;
				break;
			}
			++n;
		}
	}
	else
	{
		while( n < str.size() )
		{
			char ch = str[n];
			if( ch == ' ' ||
				ch == '\t' ||
				ch == '\r' ||
				ch == '\n' ||
				ch == '=' )
				break;

			++n;
		}
	}

	return n;
}

void CFont::InterpretKerning(string &str, int start)
{
	// Read the attributes
	int first = 0;
	int second = 0;
	int amount = 0;

	int pos, pos2 = start;
	while( true )
	{
		pos = SkipWhiteSpace(str, pos2);
		pos2 = FindEndOfToken(str, pos);

		string token = str.substr(pos, pos2-pos);

		pos = SkipWhiteSpace(str, pos2);
		if( pos == str.size() || str[pos] != '=' ) break;

		pos = SkipWhiteSpace(str, pos+1);
		pos2 = FindEndOfToken(str, pos);

		string value = str.substr(pos, pos2-pos);

		if( token == "first" )
			first = strtol(value.c_str(), 0, 10);
		else if( token == "second" )
			second = strtol(value.c_str(), 0, 10);
		else if( token == "amount" )
			amount = strtol(value.c_str(), 0, 10);

		if( pos == str.size() ) break;
	}

	// Store the attributes
	if( first >= 0 && first < 256 && chars[first] )
	{
		chars[first]->kerningPairs.push_back(second);
		chars[first]->kerningPairs.push_back(amount);
	}
}

void CFont::InterpretChar(string &str, int start)
{
	// Read all attributes
	int id = 0;
	int x = 0;
	int y = 0;
	int width = 0;
	int height = 0;
	int xoffset = 0;
	int yoffset = 0;
	int xadvance = 0;
	int page = 0;
	int chnl = 0;

	int pos, pos2 = start;
	while( true )
	{
		pos = SkipWhiteSpace(str, pos2);
		pos2 = FindEndOfToken(str, pos);

		string token = str.substr(pos, pos2-pos);

		pos = SkipWhiteSpace(str, pos2);
		if( pos == str.size() || str[pos] != '=' ) break;

		pos = SkipWhiteSpace(str, pos+1);
		pos2 = FindEndOfToken(str, pos);

		string value = str.substr(pos, pos2-pos);

		if( token == "id" )
			id = strtol(value.c_str(), 0, 10);
		else if( token == "x" )
			x = strtol(value.c_str(), 0, 10);
		else if( token == "y" )
			y = strtol(value.c_str(), 0, 10);
		else if( token == "width" )
			width = strtol(value.c_str(), 0, 10);
		else if( token == "height" )
			height = strtol(value.c_str(), 0, 10);
		else if( token == "xoffset" )
			xoffset = strtol(value.c_str(), 0, 10);
		else if( token == "yoffset" )
			yoffset = strtol(value.c_str(), 0, 10);
		else if( token == "xadvance" )
			xadvance = strtol(value.c_str(), 0, 10);
		else if( token == "page" )
			page = strtol(value.c_str(), 0, 10);
		else if( token == "chnl" )
		{
			chnl = strtol(value.c_str(), 0, 10);

			// Convert to a 4 element vector
			// TODO: Does this depend on hardware? It probably does
			if     ( chnl == 1 ) chnl = 0x00010000;  // Blue channel
			else if( chnl == 2 ) chnl = 0x00000100;  // Green channel
			else if( chnl == 4 ) chnl = 0x00000001;  // Red channel
			else if( chnl == 8 ) chnl = 0x01000000;  // Alpha channel
			else chnl = 0;
		}

		if( pos == str.size() ) break;
	}

	// Store the attributes
	if( id >= 0 && id < 256 )
	{
		if( chars[id] == 0 ) chars[id] = new SCharDescr;

		chars[id]->srcX = x;
		chars[id]->srcY = y;
		chars[id]->srcW = width;
		chars[id]->srcH = height;
		chars[id]->xOff = xoffset;
		chars[id]->yOff = yoffset;
		chars[id]->xAdv = xadvance;
		chars[id]->page = page;
		chars[id]->chnl = chnl;
	}
	else if( id == -1 )
	{
		defChar.srcX = x;
		defChar.srcY = y;
		defChar.srcW = width;
		defChar.srcH = height;
		defChar.xOff = xoffset;
		defChar.yOff = yoffset;
		defChar.xAdv = xadvance;
		defChar.page = page;
		defChar.chnl = chnl;
	}
}

void CFont::InterpretCommon(string &str, int start)
{
	// Read all attributes
	int pos, pos2 = start;
	while( true )
	{
		pos = SkipWhiteSpace(str, pos2);
		pos2 = FindEndOfToken(str, pos);

		string token = str.substr(pos, pos2-pos);

		pos = SkipWhiteSpace(str, pos2);
		if( pos == str.size() || str[pos] != '=' ) break;

		pos = SkipWhiteSpace(str, pos+1);
		pos2 = FindEndOfToken(str, pos);

		string value = str.substr(pos, pos2-pos);

		if( token == "lineHeight" )
			fontHeight = (short)strtol(value.c_str(), 0, 10);
		else if( token == "base" )
			base = (short)strtol(value.c_str(), 0, 10);
		else if( token == "scaleW" )
			scaleW = (short)strtol(value.c_str(), 0, 10);
		else if( token == "scaleH" )
			scaleH = (short)strtol(value.c_str(), 0, 10);
		else if( token == "pages" )
		{
			int pc = strtol(value.c_str(), 0, 10);
			pages.resize(pc);
			for( int n = 0; n < pc; n++ )
				pages[n] = 0;
		}

		if( pos == str.size() ) break;
	}
}

void CFont::InterpretPage(string &str, int start, const char *fontFile)
{
	int id = 0;
	string file;

	// Read all attributes
	int pos, pos2 = start;
	while( true )
	{
		pos = SkipWhiteSpace(str, pos2);
		pos2 = FindEndOfToken(str, pos);

		string token = str.substr(pos, pos2-pos);

		pos = SkipWhiteSpace(str, pos2);
		if( pos == str.size() || str[pos] != '=' ) break;

		pos = SkipWhiteSpace(str, pos+1);
		pos2 = FindEndOfToken(str, pos);

		string value = str.substr(pos, pos2-pos);

		if( token == "id" )
			id = strtol(value.c_str(), 0, 10);
		else if( token == "file" )
			file = value.substr(1, value.length()-2);

		if( pos == str.size() ) break;
	}

	// Load the texture from the same directory as the font descriptor file

	// Find the directory
	str = fontFile;
	for( size_t n = 0; (n = str.find('/', n)) != string::npos; ) str.replace(n, 1, "\\");
	size_t i = str.rfind('\\');
	if( i != string::npos )
		str = str.substr(0, i+1);
	else
		str = "";

	// Load the font textures
	str += file;
	IDirect3DTexture9 *texture = 0;
	UINT mipLevels = 0; // 0 = all
	HRESULT hr = D3DXCreateTextureFromFileEx(render->GetDevice(), str.c_str(), 
		D3DX_DEFAULT, D3DX_DEFAULT, mipLevels, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED, 
		D3DX_DEFAULT, D3DX_DEFAULT, 0, NULL, NULL, &texture);
	if( FAILED(hr) )
	{
		LOG(("Failed to load font page '%s' (%X)", str.c_str(), hr));
	}
	pages[id] = texture;
}

float CFont::AdjustForKerningPairs(int first, int second)
{	
	if( !chars[first] ) return 0;

	for( UINT n = 0; n < chars[first]->kerningPairs.size(); n += 2 )
	{
		if( chars[first]->kerningPairs[n] == second )
			return chars[first]->kerningPairs[n+1] * scale;
	}

	return 0;
}

float CFont::GetTextWidth(const char *text, int count)
{
	if( count <= 0 )
		count = (int)strlen(text);

	float x = 0;

	for( int n = 0; n < count; n++ )
	{
		SCharDescr *ch = chars[(unsigned char)text[n]];
		if( ch == 0 ) ch = &defChar;

		x += scale * (ch->xAdv);

		if( n < count-1 )
			x += AdjustForKerningPairs((unsigned char)text[n], (unsigned char)text[n+1]);
	}

	return x;
}

void CFont::SetHeight(float h)
{
	scale = h / float(fontHeight);
}

float CFont::GetHeight()
{
	return scale * float(fontHeight);
}

float CFont::GetBottomOffset()
{
	return scale * (base - fontHeight);
}

float CFont::GetTopOffset()
{
	return scale * (base - 0);
}

void CFont::InternalWrite(float x, float y, float z, const char *text, int count)
{
	int page = -1;
	render->Begin(RENDER_QUAD_LIST);

	y += scale * float(base);

	for( int n = 0; n < count; n++ )
	{
		SCharDescr *ch = chars[(unsigned char)text[n]];
		if( ch == 0 ) ch = &defChar;

		// Map the center of the texel to the corners
		// in order to get pixel perfect mapping
		float u = (float(ch->srcX)+0.5f) / scaleW;
		float v = (float(ch->srcY)+0.5f) / scaleH;
		float u2 = u + float(ch->srcW) / scaleW;
		float v2 = v + float(ch->srcH) / scaleH;

		float a = scale * float(ch->xAdv);
		float w = scale * float(ch->srcW);
		float h = scale * float(ch->srcH);
		float ox = scale * float(ch->xOff);
		float oy = scale * float(ch->yOff);

		if( ch->page != page )
		{
			render->End();
			page = ch->page;
			render->GetGraphics()->SetTexture(pages[page]);
			render->Begin(RENDER_QUAD_LIST);
		}

		render->VtxColor(0xFFFFFFFF);
		render->VtxData(ch->chnl);
		render->VtxTexCoord(u, v);
		render->VtxPos(x+ox, y-oy, z);
		render->VtxTexCoord(u2, v);
		render->VtxPos(x+w+ox, y-oy, z);
		render->VtxTexCoord(u2, v2);
		render->VtxPos(x+w+ox, y-h-oy, z);
		render->VtxTexCoord(u, v2);
		render->VtxPos(x+ox, y-h-oy, z);

		x += a;

		if( n < count-1 )
			x += AdjustForKerningPairs((unsigned char)text[n], (unsigned char)text[n+1]);
	}

	render->End();
}

void CFont::Write(float x, float y, float z, const char *text, int count, unsigned int mode)
{
	if( count <= 0 )
		count = (int)strlen(text);

	if( mode == FONT_ALIGN_CENTER )
	{
		float w = GetTextWidth(text, count);
		x -= w/2;
	}
	else if( mode == FONT_ALIGN_RIGHT )
	{
		float w = GetTextWidth(text, count);
		x -= w;
	}

	InternalWrite(x, y, z, text, count);
}

void CFont::WriteML(float x, float y, float z, const char *text, int count, unsigned int mode)
{
	// TODO: strtok skips leading delimiters,
	// We will need to split the strings into lines ourselves by
	// searching for '\n', and taking the substring up until the newline char

	if( count <= 0 )
		count = (int)strlen(text);
	
	string complete;
	complete.assign(text, count);
	complete += '\0';

	// Get first line
	char *str = strtok(&complete[0], "\n");

	while( str )
	{
		int c = (int)strlen(str);
		
		float cx = x;
		if( mode == FONT_ALIGN_CENTER )
		{
			float w = GetTextWidth(str, c);
			cx -= w/2;
		}
		else if( mode == FONT_ALIGN_RIGHT )
		{
			float w = GetTextWidth(str, c);
			cx -= w;
		}

		InternalWrite(cx, y, z, str, c);

		// Get next line
		str = strtok(0, "\n");

		y -= scale * float(fontHeight);
	}
}

void CFont::PreparePixelPerfectOutput()
{
	IDirect3DDevice9 *dev = render->GetDevice();
	CGraphics *gfx = render->GetGraphics();

	// Determine size of view port
	D3DVIEWPORT9 vp;
	dev->GetViewport(&vp);

	// Clear world matrix
	D3DXMATRIX mtx;
	D3DXMatrixIdentity(&mtx);
	gfx->SetMatrix(D3DTS_WORLD, &mtx);

	// Setup orthogonal view
	// Origin is in lower-left corner
	D3DXMatrixOrthoOffCenterLH(&mtx, 0, (float)vp.Width, 0, (float)vp.Height, vp.MinZ, vp.MaxZ);
	gfx->SetMatrix(D3DTS_VIEW, &mtx);

	D3DXMatrixIdentity(&mtx);
	gfx->SetMatrix(D3DTS_PROJECTION, &mtx);

	// Adjust the scale of the font so that the 
	// resolution of texture is the same as the screen
	scale = 1.0f;
}

} // end namespace
