#include "application.h"
#include "acframe_appwindow.h"
#include "acgfx_graphics.h"
#include "acutil_log.h"
#include "acgfx_font.h"
#include "acgfx_camera.h"
#include "acgfx_dynrender.h"
#include "acutil_timer.h"
#include <stdio.h>
#include <d3dx9.h>

using namespace acGraphics;

CApplication::CApplication() : CBaseApplication()
{
	font      = 0;
	camera    = 0;
	shadowMap = 0;
	light     = 0;
}

CApplication::~CApplication()
{
}

int CApplication::Init()
{
	int r;

	SetWindowText(wnd->GetHandle(), "Shadow");

	r = gfx->LoadEffect("data/font.fx");
	if( r < 0 )
	{
		LOG(("Failed to load 'data/font.fx' (%d)", r));
		return -1;
	}

	r = gfx->LoadEffect("data/scene.fx");
	if( r < 0 )
	{
		LOG(("Failed to load 'data/scene.fx' (%d)", r));
		return -1;
	}

	r = gfx->LoadEffect("data/copytexture.fx");
	if( r < 0 )
	{
		LOG(("Failed to load 'data/copytexture.fx' (%d)", r));
		return -1;
	}

	r = gfx->LoadEffect("data/shadowmap.fx");
	if( r < 0 )
	{
		LOG(("Failed to load 'data/shadowmap.fx' (%d)", r));
		return -1;
	}

	font = new CFont();
	r = font->Init("data/arial24.fnt", render);
	if( r < 0 )
	{
		LOG(("Failed to initialize font 'data/arial24.fnt' (%d)", r));
		return -1;
	}

	gfx->RegisterDeviceUser(this);

	camera = new CCamera(gfx);
	camera->SetPosition(D3DXVECTOR3(8, 5, 8));
	camera->SetLookAt(D3DXVECTOR3(0,0,0));
	camera->SetNearFarPlanes(0.1f, 100.0f);

	light = new CCamera(gfx);
	light->SetLookAt(D3DXVECTOR3(0,0,0));
	light->SetOrthogonal(14);
	light->SetNearFarPlanes(2.0f, 15.0f);
	light->SetAspectRatio(1);

	// Create the scene
	ID3DXMesh *mesh = 0;

	D3DXCreateBox(gfx->GetDevice(), 10, 0.1f, 10, &mesh, 0);
	meshes.push_back(mesh);
	positions.push_back(D3DXVECTOR3(0,-0.05f,0));

	D3DXCreateBox(gfx->GetDevice(), 3, 3, 3, &mesh, 0);
	meshes.push_back(mesh);
	positions.push_back(D3DXVECTOR3(0,1.5f,0));

	D3DXCreateSphere(gfx->GetDevice(), 1, 12, 8, &mesh, 0);
	meshes.push_back(mesh);
	positions.push_back(D3DXVECTOR3(3.5f, 1, .7f));

	ReinitializeDeviceObjects(gfx);

	return 0;
}

void CApplication::Uninit()
{
	ReleaseDeviceObjects();

	if( font )
	{
		delete font;
		font = 0;
	}
	
	if( light )
	{
		delete light;
		light = 0;
	}

	if( camera )
	{
		delete camera;
		camera = 0;
	}

	for( unsigned int n = 0; n < meshes.size(); n++ )
	{
		if( meshes[n] )
		{
			meshes[n]->Release();
			meshes[n] = 0;
		}
	}
}

void CApplication::Frame()
{
	// Have the light source rotate the scene
	light->SetPosition(D3DXVECTOR3(8*cos(timer->GetFrameTime()/2), 
		                           5,
								   8*sin(timer->GetFrameTime()/2)));
}

void CApplication::RenderScene(bool doShadowMap, CCamera *cam)
{
	// Render the scene
	if( doShadowMap )
	{
		gfx->SetEffect("data/shadowmap.fx");
	}
	else
	{
		gfx->SetEffect("data/scene.fx");

		// Set the variables needed for shadow map lookup
		gfx->SetEffectParam("LIGHTDIR", light->GetViewVector());
		gfx->SetTexture(shadowMap);

		// Compute the view-to-light-projection matrix, 
		// so the shadow map can be correctly applied
		D3DXMATRIX viewToLightProj = cam->GetViewMatrix();

		D3DXMatrixInverse(&viewToLightProj, 0, &viewToLightProj);
		D3DXMatrixMultiply(&viewToLightProj, &viewToLightProj, &light->GetViewMatrix());
		D3DXMatrixMultiply(&viewToLightProj, &viewToLightProj, &light->GetProjMatrix());

		gfx->SetEffectParam("VIEWTOLIGHTPROJ", viewToLightProj);
	}

	cam->SetupViewProjMtx();

	// Loop over the effect passes, and render the scene for each pass
	UINT numPasses;
	gfx->BeginEffect(&numPasses);
	for( UINT pass = 0; pass < numPasses; pass++ )
	{
		gfx->BeginEffectPass(pass);
		for( unsigned int n = 0; n < meshes.size(); n++ )
		{
			// Set the position of the mesh
			D3DXMATRIX mtx;
			D3DXMatrixTranslation(&mtx, positions[n].x, positions[n].y, positions[n].z);
			gfx->SetMatrix(D3DTS_WORLD, &mtx);
			gfx->CommitEffectChanges();

			// Render it
			meshes[n]->DrawSubset(0);
		}
		gfx->EndEffectPass();
	}
	gfx->EndEffect();
}

void CApplication::Render()
{
	IDirect3DDevice9 *dev = gfx->GetDevice();

	//-------------------
	// Generate the shadow map
	gfx->SetRenderTarget(shadowMap);
	dev->Clear(0, 0, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 0xFFFFFFFF, 1.0f, 0);
	dev->BeginScene();
	RenderScene(true, light);
	dev->EndScene();

	//------------------
	// Render the scene
	gfx->SetRenderTarget(0);
	DWORD color = D3DCOLOR_ARGB(0,123,145,145);
	dev->Clear(0, 0, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, color, 1.0f, 0);
	dev->BeginScene();
	RenderScene(false, camera);

	//------------------
	// Show the shadow map as a GUI element
	gfx->SetEffect("data/copytexture.fx");
	gfx->SetTexture(shadowMap);

	D3DXMATRIX mtx;
	D3DXMatrixIdentity(&mtx);
	gfx->SetMatrix(D3DTS_WORLD, &mtx);

	D3DVIEWPORT9 vp;
	dev->GetViewport(&vp);
	D3DXMatrixOrthoOffCenterLH(&mtx, 0, (float)vp.Width, 0, (float)vp.Height, vp.MinZ, vp.MaxZ);
	gfx->SetMatrix(D3DTS_VIEW, &mtx);

	D3DXMatrixIdentity(&mtx);
	gfx->SetMatrix(D3DTS_PROJECTION, &mtx);

	render->Begin(RENDER_QUAD_LIST);
	render->VtxTexCoord(0,0);
	render->VtxPos((float)vp.Width-138, 138, 0);
	render->VtxTexCoord(1,0);
	render->VtxPos((float)vp.Width-10, 138, 0);
	render->VtxTexCoord(1,1);
	render->VtxPos((float)vp.Width-10, 10, 0);
	render->VtxTexCoord(0,1);
	render->VtxPos((float)vp.Width-138, 10, 0);
	render->End();

	//------------------------------
	// Write the FPS
	gfx->SetEffect("data/font.fx");
	char buf[256];
	sprintf(buf, "fps: %.2f", timer->GetFps());
	font->PreparePixelPerfectOutput();
	font->Write(5, 5, 0, buf, 0, 0);

	// Write the info
	int h;
	gfx->GetScreenDimensions(0, &h);
	font->Write(5, h-font->GetHeight(), 0, "Shadow test", 0, 0);

	dev->EndScene();
}

int CApplication::InitializeDeviceObjects()
{
	int r;

	// Create the render target for the shadow map
	r = gfx->CreateRenderTargetForShadowMap(512, 512, &shadowMap);
	if( r < 0 ) return r;

	return 0;
}

void CApplication::ReinitializeDeviceObjects(CGraphics *gfx)
{
	InitializeDeviceObjects();
}

void CApplication::ReleaseDeviceObjects()
{
	if( shadowMap )
	{
		shadowMap->Release();
		shadowMap = 0;
	}
}
