/*

    This effect will render a scene lit with a parallel light and using a shadow map

*/

shared float4x4 g_mWorld           : WORLD;           // World matrix
shared float4x4 g_mView            : VIEW;            // View matrix
shared float4x4 g_mProj            : PROJECTION;      // Projection matrix
shared float4x4 g_mViewToLightProj : VIEWTOLIGHTPROJ; // View to light projection matrix

texture  g_txShadowMap     : TEXTURE0;
float3   g_vLightDirection : LIGHTDIR; // The directional light's direction

#define SHADOW_BIAS 0.0005f
#define SHADOW_SIZE 512

sampler2D g_samShadow =
sampler_state
{
    Texture = <g_txShadowMap>;
    MinFilter = Point;
    MagFilter = Point;
    MipFilter = Point;
    AddressU = Clamp;
    AddressV = Clamp;
};

struct VS_OUTPUT
{
    float4 Position   : POSITION;   // vertex position 
    float4 Diffuse    : COLOR0;     // lit vertex color
    float4 vPos       : TEXCOORD0;  // vertex position
    float4 vPosLS     : TEXCOORD1;  // position in the light's space
    float3 vNormal    : TEXCOORD2;  // normal
};
VS_OUTPUT VertScene( float4 vPos         : POSITION,
                     float3 vNormal      : NORMAL )
{
    VS_OUTPUT Out;

    // Determine the position in view space, this will be passed on to the pixel shader
    Out.vPos = mul( vPos, g_mWorld );
    Out.vPos = mul( Out.vPos, g_mView );
    
    // Project the position onto the screen, for the pipeline
    Out.Position = mul( Out.vPos, g_mProj );

    // Compute the degree of lighting

    // Transform the normal from object space to world space
    Out.vNormal = normalize(mul(vNormal, (float3x3)g_mWorld));

    // The light direction is in the opposite direction of the normal
    float val = max(0, -dot(normalize(g_vLightDirection), Out.vNormal));
    Out.Diffuse = float4(val,val,val,1);

    // Transform the position from view space into the light's projection space
    Out.vPosLS = mul( Out.vPos, g_mViewToLightProj );

    return Out;
}


float4 PixScene( VS_OUTPUT In ) : COLOR0
{
    float4 color = 1;
    
    // Is the pixel facing the light?
    if( dot(g_vLightDirection, In.vNormal) < 0 )
    {
        float LightAmount;
    
        // Is the pixel lit?
/*
        // Without 2x2 percentage filtering
        // Gives aliased edges
        {
            // Transform the position in light space to texture space
            float2 ShadowTexC = 0.5 * In.vPosLS.xy / In.vPosLS.w + float2( 0.5, 0.5 );
            ShadowTexC.y = 1.0f - ShadowTexC.y;
            
            // Is the pixel lit?
            LightAmount = ((tex2D( g_samShadow, ShadowTexC ) + SHADOW_BIAS) < In.vPosLS.z / In.vPosLS.w)? 0.0f: 1.0f; 
        }
*/
        // With 2x2 percentage filtering
        // Gives smoother edges, like bilinear filtering
        {
            // Transform the position in light space to texture space
            float2 ShadowTexC = 0.5 * In.vPosLS.xy / In.vPosLS.w + float2( 0.5, 0.5 );
            ShadowTexC.y = 1.0f - ShadowTexC.y;
            
            // transform to texel space
            float2 texelpos = SHADOW_SIZE * ShadowTexC;
            // Determine the lerp amounts           
            float2 lerps = frac( texelpos );
    
            //read in bilerp stamp, doing the shadow checks
            float sourcevals[4];
            sourcevals[0] = (tex2D( g_samShadow, ShadowTexC ) + SHADOW_BIAS < In.vPosLS.z / In.vPosLS.w)? 0.0f: 1.0f;  
            sourcevals[1] = (tex2D( g_samShadow, ShadowTexC + float2(1.0/SHADOW_SIZE, 0) ) + SHADOW_BIAS < In.vPosLS.z / In.vPosLS.w)? 0.0f: 1.0f;  
            sourcevals[2] = (tex2D( g_samShadow, ShadowTexC + float2(0, 1.0/SHADOW_SIZE) ) + SHADOW_BIAS < In.vPosLS.z / In.vPosLS.w)? 0.0f: 1.0f;  
            sourcevals[3] = (tex2D( g_samShadow, ShadowTexC + float2(1.0/SHADOW_SIZE, 1.0/SHADOW_SIZE) ) + SHADOW_BIAS < In.vPosLS.z / In.vPosLS.w)? 0.0f: 1.0f;  
            
            // lerp between the shadow values to calculate our light amount
            LightAmount = lerp( lerp( sourcevals[0], sourcevals[1], lerps.x ),
                                lerp( sourcevals[2], sourcevals[3], lerps.x ),
                                lerps.y );
        }
        
        // Multiply with interpolated diffuse light
        color.rgb = (In.Diffuse.rgb * LightAmount);
    }
    else
    {
        // Complete shadow
        color = float4(0,0,0,1);
    }

    return color;
}


//--------------------------------------------------------------------------------------
// Techniques
//--------------------------------------------------------------------------------------
technique RenderScene
{
    pass P0
    {
		// Turn on alpha blending
		AlphaBlendEnable = true;
		DestBlend        = INVSRCALPHA;
		SrcBlend         = SRCALPHA;
		
		// Set the vertex and pixel shaders
        VertexShader = compile vs_2_0 VertScene();
        PixelShader  = compile ps_2_0 PixScene();
    }
}
